/*
    Copyright (C) 1996-2001  Ulric Eriksson <ulric@siag.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "common.h"
#include <Mowitz/MwUtils.h>
#include "oldfonts.h"
#include "fonts.h"

siag_format format_table[256];
int nformat;

struct fa_struct fontalias[256];

int nfontalias = 0;

static char *glyph_names[256];
siag_fontname fontname_table[256];
int nfontname;

static struct {
	char *name;
	int mask;
} attrnames[] = {
	{"family", SFMT_FAMILY},
	{"size", SFMT_SIZE},
	{"fg", SFMT_FG},
	{"bg", SFMT_BG},
	{"bold", SFMT_BOLD},
	{"italic", SFMT_ITALIC},
	{"uline", SFMT_ULINE},
	{"strike", SFMT_STRIKE},
	{"hadj", SFMT_HADJ},
	{"vadj", SFMT_VADJ},
	{"borders", SFMT_BORDERS},
	{"style", SFMT_STYLE},
	{NULL, 0}
};

/* ---
Return an old-style attribute mask from its name.
*/

int fmt_attr2mask(char *attr)
{
	int i;

	for (i = 0; attrnames[i].name; i++) {
		if (!MwStrcasecmp(attr, attrnames[i].name))
			return attrnames[i].mask;
	}
	return 0;
}

/* ---
Return an attribute name from its old-style mask.
*/

char *fmt_mask2attr(int mask)
{
	int i;

	for (i = 0; attrnames[i].name; i++) {
		if (mask == attrnames[i].mask)
			return attrnames[i].name;
	}
	return NULL;
}


/* ---
returns the index of a format structure with all the requested
   properties, or -1 for failure. It is assumed that the font has been
   allocated by calling lookup_font and the colors by calling
   lookup_color. lookup_format doesn't allocate any X resources.

   981106: Added "style", which is a bit tricky because we have no
   idea in this code what a style is or does: it varies between
   applications. The application encodes the style in to an integer
   and we don't care what it really means.
*/

int lookup_format(int font, int uline, int strike, int fgcolor, int bgcolor,
		int style, int siagfmt)
{
	int i;

	for (i = 0; i < nformat; i++) {
		if (format_table[i].font == font &&
			format_table[i].uline == uline &&
			format_table[i].strike == strike &&
			format_table[i].fg == fgcolor &&
			format_table[i].bg == bgcolor &&
			format_table[i].style == style &&
			format_table[i].siagfmt == siagfmt)
				return i;
	}
	nformat++;
	format_table[i].font = font;
	format_table[i].uline = uline;
	format_table[i].strike = strike;
	format_table[i].fg = fgcolor;
	format_table[i].bg = bgcolor;
	format_table[i].style = style;
	format_table[i].siagfmt = siagfmt;
	return i;
}

/* ---
Loading and saving formats.

This code is shared between all programs that use the new formats.
That way there will be less duplication of code.

This requires new tags to store the formats. The new tag is

.ft N

where N is a number. This is followed by additional lines of
the format

name value

where name is the name of an attribute and
value is the value of the attribute. The beauty of this is
that it can be extended with new attributes without breaking
compatibility. Old versions of the program will simply not
understand new attributes and will therefore ignore them.
A format specification is terminated by the word "end" alone
on a line.

Example:

.ft 0
family Times
size 10
uline n
fg red
end
.ft 1
family Courier
size 12
italic y
end

It is not necessary to specify all attributes. The ones that are
not specified will use defaults compiled into the program.

Compatibility: Old files contain no .ft tags. For those, a function
will be used that translates from the old bitmap fields to new
structures. Another function translates in the other direction,
making it possible to store in a format old versions of Siag
can open.
--- */

/* ---
Convert an old-style bitmapped format to the new style index.
*/

int fmt_old2new(long oldfmt)
{
	sfmt fmt;

	switch (oldfmt & SIZE_MASK) {
	case SIZE_8: fmt.size = 80; break;
	case SIZE_10: fmt.size = 100; break;
	case SIZE_12: fmt.size = 120; break;
	case SIZE_14: fmt.size = 140; break;
	case SIZE_18: fmt.size = 180; break;
	case SIZE_24: fmt.size = 240; break;
	case SIZE_20: fmt.size = 200; break;
	default: fmt.size = 300; break;
	}
	fmt.italic = ((oldfmt & ITALIC) != 0);
	fmt.bold = ((oldfmt & BOLD) != 0);
	fmt.uline = ((oldfmt & ULINE) != 0);
	fmt.strike = 0;		/* strikethrough didn't exist then */
	switch (oldfmt & FONT_MASK) {
	case COURIER: fmt.family = "Courier"; break;
	case HELVETICA: fmt.family = "Helvetica"; break;
	case NEW_CENTURY: fmt.family = "New Century Schoolbook"; break;
	default: fmt.family = "Times"; break;
	}
	switch (oldfmt & COLOR_MASK) {
	case COLOR_0: fmt.fg = "black"; break;
	case COLOR_1: fmt.fg = "red"; break;
	case COLOR_2: fmt.fg = "green"; break;
	case COLOR_3: fmt.fg = "blue"; break;
	case COLOR_4: fmt.fg = "yellow"; break;
	case COLOR_5: fmt.fg = "magenta"; break;
	case COLOR_6: fmt.fg = "cyan"; break;
	default: fmt.fg = "white"; break;
	}
	fmt.bg = "white";
	fmt.borders = oldfmt & BORDER_MASK;
	fmt.vadj = oldfmt & VADJ_MASK;
	fmt.hadj = oldfmt & HADJ_MASK;
	/* this works because the old styles are registered first */
	fmt.style = (oldfmt & FMT_MASK) >> FMT_SHIFT;
	return encode_format(~0, &fmt);
}

/* ---
Convert a new-style format index to the old bitmapped format.
this function could try harder
*/

long fmt_new2old(int newfmt)
{
	return SIZE_10 | HELVETICA;
}

/* the default format */
static sfmt dfmt = { "Helvetica", 120, 0, 0, 0, 0, "black", "white",
			0, VADJ_CENTER, HADJ_LEFT, 0};

/* ---
the formats are not split up into fonts, colors and rest, but saved
   in a single level. The restructuring is done when the formats are read.
   Attributes: font, size, bold, italic, uline, strike, fg, bg, siagfmt
*/

void save_formats(FILE *fp, int i)
{
	sfmt fmt;

	decode_format(i, ~0, &fmt);
	fprintf(fp, ".ft %d\n", i);
	if (MwStrcasecmp(fmt.family, dfmt.family))
		fprintf(fp, "font %s\n", fmt.family);
	if (fmt.size != dfmt.size)
		fprintf(fp, "size %d\n", fmt.size);
	if (fmt.bold != dfmt.bold)
		fprintf(fp, "bold %d\n", fmt.bold);
	if (fmt.italic != dfmt.italic)
		fprintf(fp, "italic %d\n", fmt.italic);
	if (fmt.uline != dfmt.uline)
		fprintf(fp, "uline %d\n", fmt.uline);
	if (fmt.strike != dfmt.strike)
		fprintf(fp, "strike %d\n", fmt.strike);
	if (MwStrcasecmp(fmt.fg, dfmt.fg))
		fprintf(fp, "fg %s\n", fmt.fg);
	if (MwStrcasecmp(fmt.bg, dfmt.bg))
		fprintf(fp, "bg %s\n", fmt.bg);
	if (fmt.borders != dfmt.borders)
		fprintf(fp, "borders %d\n", fmt.borders);
	if (fmt.vadj != dfmt.vadj)
		fprintf(fp, "vadj %d\n", fmt.vadj);
	if (fmt.hadj != dfmt.hadj)
		fprintf(fp, "hadj %d\n", fmt.hadj);
	if (fmt.style != dfmt.style)
		fprintf(fp, "style %d\n", fmt.style);
	fprintf(fp, "end\n");
}

/* ---
When this function gets to borrow the file pointer, we are about to
   read the first attribute of a font. We keep on reading until eof
   or "end". The value returned is the index into the real format table.
*/

int load_formats(FILE *fp)
{
	char b[1000];
	char font[256], fg[256], bg[256];
	sfmt fmt = dfmt;

	while (fgets(b, sizeof b, fp) && strncmp(b, "end", 3)) {
		MwChomp(b);
		if (!strncmp(b, "font ", 5)) {
			strcpy(font, b+5);
			fmt.family = font;
		} else if (!strncmp(b, "size ", 5)) {
			fmt.size = strtol(b+5, NULL, 10);
		} else if (!strncmp(b, "bold ", 5)) {
			fmt.bold = strtol(b+5, NULL, 10);
		} else if (!strncmp(b, "italic ", 7)) {
			fmt.italic = strtol(b+7, NULL, 10);
		} else if (!strncmp(b, "uline ", 6)) {
			fmt.uline = strtol(b+6, NULL, 10);
		} else if (!strncmp(b, "strike ", 6)) {
			fmt.strike = strtol(b+7, NULL, 10);
		} else if (!strncmp(b, "fg ", 3)) {
			strcpy(fg, b+3);
			fmt.fg = fg;
		} else if (!strncmp(b, "bg ", 3)) {
			strcpy(bg, b+3);
			fmt.bg = bg;
		} else if (!strncmp(b, "borders ", 8)) {
			fmt.borders = strtol(b+8, NULL, 10);
		} else if (!strncmp(b, "vadj ", 5)) {
			fmt.vadj = strtol(b+5, NULL, 10);
		} else if (!strncmp(b, "hadj ", 5)) {
			fmt.hadj = strtol(b+5, NULL, 10);
		} else if (!strncmp(b, "style ", 6)) {
			fmt.style = strtol(b+6, NULL, 10);
		}
		/* ignore anything we don't grok */
	}
	return encode_format(~0, &fmt);
}

static int tmpncolor;
static struct colors {
	int r, g, b;
	char *n;
} *tmpcolor;

static void scan_name(char *from, char *to)
{
	int c = ' ';
	while (*from) {
		if (isspace(c) && islower(*from)) {
			*to++ = c = toupper(*from++);
		} else if (isupper(*from)) {
			if (!isspace(c)) *to++ = ' ';
			*to++ = c = *from++;
		} else {
			*to++ = c = *from++;
		}
	}
	*to = '\0';
}

static void insert_color(int r, int g, int b, char *n)
{
	int i;
	char n2[1000];
	scan_name(n, n2);
	for (i = 0; i < tmpncolor; i++) {
		if (!strcmp(n2, tmpcolor[i].n)) break;
	}
	if (i == tmpncolor) {
		tmpncolor++;
		tmpcolor = realloc(tmpcolor, tmpncolor*sizeof *tmpcolor);
	} else {
		free(tmpcolor[i].n);
	}
	tmpcolor[i].r = r;
	tmpcolor[i].g = g;
	tmpcolor[i].b = b;
	tmpcolor[i].n = malloc(strlen(n2)+1);
	strcpy(tmpcolor[i].n, n2);
}

static int ccompar(const void *p, const void *q)
{
	const struct colors *c1 = p;
	const struct colors *c2 = q;
	return strcmp(c1->n, c2->n);
}

static int sc(int n)
{
	return n*65535/255;
}

void init_colors(void)
{
	char fn[1000], s[1000], n[1000];
	FILE *fp;
	int r, g, b, i;

	sprintf(fn, "%s/common/rgb.txt", datadir);
	fp = fopen(fn, "r");
	if (fp == NULL) {
		fprintf(stderr, "Can't read %s\n", fn);
		return;
	}

	while (fgets(s, sizeof s, fp)) {
		if (s[0] != '!') {
			i = sscanf(s, "%d %d %d %[^\n]", &r, &g, &b, n);
			if (i == 4) {
				insert_color(r, g, b, n);
			}
		}
	}
	fclose(fp);
	qsort(tmpcolor, tmpncolor, sizeof *tmpcolor, ccompar);
	for (i = 0; i < tmpncolor; i++) {
		register_color(tmpcolor[i].n, sc(tmpcolor[i].r),
			sc(tmpcolor[i].g), sc(tmpcolor[i].b));
	}
}

static void load_encoding(char *fn)
{
	int state = 0;
	char b[256];
	FILE *fp = fopen(fn, "r");
	int i, n = 0;
	if (fp == NULL) return;

	while (fgets(b, sizeof b, fp)) {
		if (!strncmp(b, "Encoding=", 9)) {
                        state = 1;
                } else if (state) {
                        for (i = 0; b[i] && !isspace(b[i]); i++);
                        b[i] = '\0';
                        glyph_names[n++] = MwStrdup(b);
                }
        }
        fclose(fp);
}

static int find_glyph(char *n)
{
	int i;
 
        for (i = 0; i < 256; i++)
                if (glyph_names[i] && !strcmp(glyph_names[i], n)) return i;
        return -1;
}

int *load_afm(char *q)
{
	char fn[1024];
	int state = 0, w, g;
	char b[256], n[256], *p;
	int *widths = MwCalloc(512, sizeof *widths);
	FILE *fp;

	if (q[0] == '/') strcpy(fn, q);
	else sprintf(fn, "%s/common/fonts/%s", datadir, q);

	fp = fopen(fn, "r");
	if (fp == NULL) return NULL;

	while (fgets(b, sizeof b, fp)) {
                if (!strncmp(b, "StartCharMetrics", 16)) {
                        state = 1;
                } else if (!strncmp(b, "EndCharMetrics", 14)) {
                        state = 0;
                } else if (state) {
                        p = strstr(b, "WX ");
                        if (!p) continue;
                        sscanf(p, "WX %d", &w);
                        p = strstr(b, "C ");
                        if (p && sscanf(p, "C %d", &g) == 1 && g >= 0)
                                widths[g] = w;
                        p = strstr(b, "N ");
                        if (!p) continue;
                        sscanf(p, "N %s", n);
                        g = find_glyph(n);
                        if (g == -1) continue;
                        widths[g] = w;
                }
        }
	fclose(fp);
	return widths;
}

static void grow_fontname_table(char *name)
{
	fontname_table[nfontname].name = MwStrdup(name);
	fontname_table[nfontname].afm[0] = NULL;
	fontname_table[nfontname].afm[1] = NULL;
	fontname_table[nfontname].afm[2] = NULL;
	fontname_table[nfontname].afm[3] = NULL;
	fontname_table[nfontname].pfb[0] = NULL;
	fontname_table[nfontname].pfb[1] = NULL;
	fontname_table[nfontname].pfb[2] = NULL;
	fontname_table[nfontname].pfb[3] = NULL;
	fontname_table[nfontname].x_name[0] = NULL;
	fontname_table[nfontname].x_name[1] = NULL;
	fontname_table[nfontname].x_name[2] = NULL;
	fontname_table[nfontname].x_name[3] = NULL;
	fontname_table[nfontname].ps_name[0] = NULL;
	fontname_table[nfontname].ps_name[1] = NULL;
	fontname_table[nfontname].ps_name[2] = NULL;
	fontname_table[nfontname].ps_name[3] = NULL;
	fontname_table[nfontname].t1_name[0] = NULL;
	fontname_table[nfontname].t1_name[1] = NULL;
	fontname_table[nfontname].t1_name[2] = NULL;
	fontname_table[nfontname].t1_name[3] = NULL;
	nfontname++;
}

/* ---
X, T1 and PS fonts are registered separately because they are device
dependent on different devices. This is slightly lame because they will
probably be called from the same Scheme wrapper function, and they are
in the same C translation unit. Never mind.
*/

int register_afm(char *name, char *a, char *ai, char *ab, char *abi)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!strcasecmp(name, fontname_table[i].name)) break;
	}
	if (i == nfontname) grow_fontname_table(name);
	fontname_table[i].afm[0] = load_afm(a);
	fontname_table[i].afm[1] = load_afm(ai);
	fontname_table[i].afm[2] = load_afm(ab);
	fontname_table[i].afm[3] = load_afm(abi);
	return i;
}

int register_x_font(char *name, char *x, char *xi, char *xb, char *xbi)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!strcasecmp(name, fontname_table[i].name)) break;
	}
	if (i == nfontname) grow_fontname_table(name);
	fontname_table[i].x_name[0] = MwStrdup(x);
	fontname_table[i].x_name[1] = MwStrdup(xi);
	fontname_table[i].x_name[2] = MwStrdup(xb);
	fontname_table[i].x_name[3] = MwStrdup(xbi);
	return i;
}

static void register_encoding(char *name)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!MwStrcasecmp(name, fontname_table[i].name)) break;
	}
	if (i < nfontname) fontname_table[i].iso8859_1 = 0;
}

/* ---
As above but for Postscript.
*/

int register_ps_font(char *name, char *ps, char *psi, char *psb, char *psbi)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!MwStrcasecmp(name, fontname_table[i].name)) break;
	}
	if (i == nfontname) grow_fontname_table(name);
	fontname_table[i].ps_name[0] = MwStrdup(ps);
	fontname_table[i].ps_name[1] = MwStrdup(psi);
	fontname_table[i].ps_name[2] = MwStrdup(psb);
	fontname_table[i].ps_name[3] = MwStrdup(psbi);
	/* and now, a piece written in Hackish */
	fontname_table[i].iso8859_1 = strcmp(ps, psi)?1:0;
	return i;
}

/* ---
As above but for pfb files.
*/

int register_pfb(char *name, char *ps, char *psi, char *psb, char *psbi)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!MwStrcasecmp(name, fontname_table[i].name)) break;
	}
	if (i == nfontname) grow_fontname_table(name);
	fontname_table[i].pfb[0] = MwStrdup(ps);
	fontname_table[i].pfb[1] = MwStrdup(psi);
	fontname_table[i].pfb[2] = MwStrdup(psb);
	fontname_table[i].pfb[3] = MwStrdup(psbi);
	fontname_table[i].iso8859_1 = 0;
	return i;
}

/* ---
As above but for Type 1 (T1lib).
*/

int register_t1_font(char *name, char *ps, char *psi, char *psb, char *psbi)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!MwStrcasecmp(name, fontname_table[i].name)) break;
	}
	if (i == nfontname) grow_fontname_table(name);
	fontname_table[i].t1_name[0] = MwStrdup(ps);
	fontname_table[i].t1_name[1] = MwStrdup(psi);
	fontname_table[i].t1_name[2] = MwStrdup(psb);
	fontname_table[i].t1_name[3] = MwStrdup(psbi);
	/* and now, a piece written in Hackish */
	fontname_table[i].iso8859_1 = strcmp(ps, psi)?1:0;
	return i;
}

/* ---
handle mapping between e.g. Arial and Helvetica

Store the name of a font we don't have and the name of a replacement
*/

int font_alias(char *alias, char *name)
{
	int i;

	for (i = 0; i < nfontalias; i++) {
		if (!MwStrcasecmp(alias, fontalias[i].alias)) break;
	}
	if (i == nfontalias) {
		nfontalias++;
		fontalias[i].alias = MwStrdup(alias);
	}
	fontalias[i].name = MwStrdup(name);
	return i;
}

/* ---
returns the index of a font with the specified name, or -1 for failure
*/

int lookup_fontname(char *name)
{
	int i;

	for (i = 0; i < nfontname; i++) {
		if (!MwStrcasecmp(name, fontname_table[i].name))
			return i;
	}
	return -1;
}


/* ---
If there is a font with exactly the alias name, return that. Otherwise
if there is an alias, return the name of the real font. Otherwise NULL.
*/

char *lookup_fontalias(char *alias)
{
	int i;

	if (alias == NULL) return NULL;
	if (lookup_fontname(alias) >= 0) return alias;
	for (i = 0; i < nfontalias; i++) {
		if (!MwStrcasecmp(alias, fontalias[i].alias))
			return fontalias[i].name;
	}
	return NULL;
}

void init_fonts(void)
{
	char fn[1000], s[1000];
	char n[1000], nr[1000], ni[1000], nb[1000], nbi[1000];
	int i;
	FILE *fp;

	sprintf(fn, "%s/common/IsoLatin1.enc", datadir);
	load_encoding(fn);

	sprintf(fn, "%s/common/fonts.txt", datadir);
	fp = fopen(fn, "r");
	if (fp == NULL) {
		fprintf(stderr, "Can't read %s\n", fn);
		return;
	}

	n[0] = '\0';
	while (fgets(s, sizeof s, fp)) {
		if (s[0] == '[') {
			i = sscanf(s, "[%[^]]", n);
			if (i != 1) {
				fprintf(stderr, "Error in %s\n", s);
			}
		} else if (!strncmp(s, "AFM=", 4)) {
			i = sscanf(s, "AFM=%[^|]|%[^|]|%[^|]|%[^\n]",
				nr, ni, nb, nbi);
			if (i != 4) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				register_afm(n, nr, ni, nb, nbi);
			}
		} else if (!strncmp(s, "PFB=", 4)) {
			i = sscanf(s, "PFB=%[^|]|%[^|]|%[^|]|%[^\n]",
				nr, ni, nb, nbi);
			if (i != 4) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				register_pfb(n, nr, ni, nb, nbi);
			}
		} else if (!strncmp(s, "X=", 2)) {
			i = sscanf(s, "X=%[^|]|%[^|]|%[^|]|%[^\n]",
				nr, ni, nb, nbi);
			if (i != 4) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				register_x_font(n, nr, ni, nb, nbi);
			}
		} else if (!strncmp(s, "PS=", 3)) {
			i = sscanf(s, "PS=%[^|]|%[^|]|%[^|]|%[^\n]",
				nr, ni, nb, nbi);
			if (i != 4) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				register_ps_font(n, nr, ni, nb, nbi);
			}
		} else if (!strncmp(s, "T1=", 3)) {
			i = sscanf(s, "T1=%[^|]|%[^|]|%[^|]|%[^\n]",
				nr, ni, nb, nbi);
			if (i != 4) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				register_t1_font(n, nr, ni, nb, nbi);
			}
		} else if (!strncmp(s, "ALIAS=", 6)) {
			i = sscanf(s, "ALIAS=%[^\n]", nr);
			if (i != 1) {
				fprintf(stderr, "Error in %s\n", s);
			} else {
				font_alias(nr, n);
			}
		} else if (!strncmp(s, "ENC=", 4)) {
			/* tell us it's not Latin1 */
			register_encoding(n);
		}
	}
	fclose(fp);
}

