/*
    Copyright (C) 1997-2000  Ulric Eriksson <ulric@siag.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

/*
 * fonts.c
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "../common/common.h"
#include <Mowitz/MwUtils.h>
#include "../common/oldfonts.h"
#include "../common/fonts.h"
#include "fonts.h"

siag_color color_table[1000];
siag_font font_table[256];
static int ncolor, nfontname, nfont;

static int nfontalias = 0;

static int lastfont;

static int compar(const void *p, const void *q)
{
	return strcmp(*(const char **)p, *(const char **)q);
}

/* ---
returns a null-terminated, sorted list of all known fonts.
   Caller must free
*/

char **font_list(int *n)
{
	int i, j;
	char **list = (char **)MwMalloc((nfontalias+nfontname+1)*sizeof(char *));
	for (i = 0; i < nfontalias; i++)
		list[i] = MwStrdup(fontalias[i].alias);
	for (j = 0; j < nfontname; j++)
		list[i+j] = MwStrdup(fontname_table[j].name);
	list[i+j] = NULL;
	qsort(list, i+j, sizeof(char *), compar);
	*n = i+j;
	return list;
}

/* ---
returns a null-terminated, sorted list of all known colours.
   Caller must free
*/

char **color_list(int *n)
{
	int i;
	char **list = (char **)MwMalloc((ncolor+1)*sizeof(char *));
	for (i = 0; i < ncolor; i++)
		list[i] = MwStrdup(color_table[i].name);
	list[i] = NULL;
	qsort(list, i, sizeof(char *), compar);
	*n = i;
	return list;
}


/* ---
Take an index into the format table (i.e. a new-style format code) and
decode it into the components specified by the mask.
*/

void decode_format(int n, long mask, sfmt *fmt)
{
	int font = format_table[n].font;

	if (mask & SFMT_FAMILY) {
		fmt->family = fontname_table[font_table[font].name].name;
		if (!fmt->family) fmt->family = "Helvetica";
	}
	if (mask & SFMT_SIZE)
		fmt->size = font_table[font].size;
	if (mask & SFMT_BOLD)
		fmt->bold = font_table[font].bold;
	if (mask & SFMT_ITALIC)
		fmt->italic = font_table[font].italic;
	if (mask & SFMT_ULINE)
		fmt->uline = format_table[n].uline;
	if (mask & SFMT_FG) {
		fmt->fg = color_table[format_table[n].fg].name;
		if (!fmt->fg) fmt->fg = "black";
	}
	if (mask & SFMT_BG) {
		fmt->bg = color_table[format_table[n].bg].name;
		if (!fmt->bg) fmt->bg = "white";
	}
	if (mask & SFMT_BORDERS)
		fmt->borders = format_table[n].siagfmt & BORDER_MASK;
	if (mask & SFMT_VADJ)
		fmt->vadj = format_table[n].siagfmt & VADJ_MASK;
	if (mask & SFMT_HADJ)
		fmt->hadj = format_table[n].siagfmt & HADJ_MASK;
	if (mask & SFMT_STYLE)
		fmt->style = format_table[n].style;
}

/* ---
Encode the components from fmt specified by the mask into a new-style
format code, allocating colours and fonts as necessary.
*/

int encode_format(long mask, sfmt *fmt)
{
	char *family = "Helvetica";
	int fg = 0, bg = 0;
	int size = 120;
	int bold = 0, italic = 0, uline = 0, strike = 0;
	int siagfmt = 0;
	int style = 0;

	if (mask & SFMT_FAMILY) {
		family = lookup_fontalias(fmt->family);
		if (!family) family = "Helvetica";
	}
	if (mask & SFMT_SIZE)
		size = fmt->size;
	if (mask & SFMT_BOLD)
		bold = fmt->bold;
	if (mask & SFMT_ITALIC)
		italic = fmt->italic;
	if (mask & SFMT_ULINE)
		uline = fmt->uline;
	if (mask & SFMT_STRIKE)
		strike = fmt->strike;
	if (mask & SFMT_FG) {
		fg = lookup_color(fmt->fg);
		if (fg == -1) fg = 0;	/* black */
	}
	if (mask & SFMT_BG) {
		bg = lookup_color(fmt->bg);
		if (bg == -1) bg = 7;	/* white */
	}
	if (mask & SFMT_BORDERS)
		siagfmt |= fmt->borders;
	if (mask & SFMT_VADJ)
		siagfmt |= fmt->vadj;
	if (mask & SFMT_HADJ)
		siagfmt |= fmt->hadj;
	if (mask & SFMT_STYLE)
		style = fmt->style;
	return lookup_format(lookup_font(family, size, bold, italic),
				uline, strike, fg, bg, style, siagfmt);
}

/* ---
Here is the whole song and dance to get a format id for Siag versions >= 3.1.

Start by deciding on font family, point size, weight and slant.
Get an index into the font table by calling lookup_font with
the family name as argument along with the size, weight and slant.
The font is automatically loaded, if necessary.

Next pick foreground and background colours. The colours are loaded
by calling lookup_color with the colour name as argument. An index into
the colour table is returned.

Finally we are ready to call lookup_format. Decide if the character
should be underlined. If the application is Siag, also combine the
cell attributes. Pass the font index, the colour index, the underlining
flag and the Siag attributes to lookup_format. An index into the
format table is returned, and it is this index we can use to replace
the old style formats.

Yes, this is more complex than the old stuff. But it is also in fact
easier to extract individual pieces of the formats, since they are
normal C structures. To my eyes, font_table[format_table[fmt].font].size
is much clearer than (fmt & SIZE_MASK)>>SIZE_SHIFT. And it is less
prone to errors. And it leaves room for future expansion, for example
for overstrike and double underline, or bidirectional printing.
--- */

/* ---
returns the index of a structure with the requested color, or -1
   for failure. Pixel value in color_table[i].color.pixel is allocated
*/

int lookup_color(char *name)
{
	int i;

	if (!name) {
		fprintf(stderr, "Some dumbass called me with null colour!\n");
		abort();
	}

	for (i = 0; i < ncolor; i++) {
		if (!MwStrcasecmp(name, color_table[i].name))
			break;
	}
	if (i == ncolor) return -1;	/* fail */
	return i;
}

/* ---
return index into font table or -1 for failure. Load font if necessary.
*/

int lookup_font(char *fname, int size, int bold, int italic)
{
	int i, name;

	name = lookup_fontname(fname);
	if (name == -1) name = 0;	/* fallback value */

	for (i = 0; i < nfont; i++) {
		if (font_table[i].name == name &&
			font_table[i].size == size &&
			font_table[i].bold == bold &&
			font_table[i].italic == italic)
				return i;	/* already loaded */
	}
	nfont++;
	font_table[i].name = name;
	font_table[i].size = size;
	font_table[i].bold = bold;
	font_table[i].italic = italic;
	return i;
}

/* ---
stores a colour name along with the red, green and blue values,
   or overwrites an existing colour with the same name.
   The colour is not allocated until it is looked up in lookup_color.
*/

int register_color(char *name,
	unsigned short red, unsigned short green, unsigned short blue)
{
	int i;

	for (i = 0; i < ncolor; i++)
		if (!MwStrcasecmp(name, color_table[i].name)) break;
	if (i == ncolor) {
		ncolor++;
		color_table[i].name = MwStrdup(name);
	}
	color_table[i].red = red;
	color_table[i].green = green;
	color_table[i].blue = blue;
	color_table[i].need_init = 1;
	return i;
}

/* ---
fmt is an index into fontname_table
*/

char *family2name(int fmt)
{
	return fontname_table[fmt].name;
}

/* ---
fmt is index into font_table
*/

char *size2name(int fmt)
{
	static char b[80];

	sprintf(b, "%d", font_table[fmt].size/10);
	return b;
}

/* ---
fmt is index into color_table
*/

char *color2name(int fmt)
{
	return color_table[fmt].name;
}

/* ---
Postscript font name
*/

char *ps_fontname(int fmt)
{
	return font_table[format_table[fmt].font].ps_name;
}

/* ---
Postscript font size
*/

int ps_fontsize(int fmt)
{
	return font_table[format_table[fmt].font].size / 10;
}

/* ---
Define Latin-1 fonts
*/

void ps_makefonts(FILE *fp)
{
	int i, j;

	for (i = 0; i < nfontname; i++) {
		if (!fontname_table[i].iso8859_1)
			continue;	/* don't bother */
		for (j = 0; j < 3; j++) {
			if (j && !strcmp(fontname_table[i].ps_name[0],
					fontname_table[i].ps_name[j]))
				continue;	/* done already */
			fprintf(fp, "/%s findfont\n",
				fontname_table[i].ps_name[j]);

			fprintf(fp, "dup length dict begin\n");
			fprintf(fp, "  {1 index /FID ne {def} {pop pop} ifelse} forall\n");
			fprintf(fp, "  /Encoding ISOLatin1Encoding def\n");
			fprintf(fp, "  currentdict\n");
			fprintf(fp, "end\n");
			fprintf(fp, "/%s-ISOLatin1 exch definefont pop\n",
				fontname_table[i].ps_name[j]);
		}
	}
}

void ps_set_color(FILE *fp, int red, int green, int blue)
{
	;
}

/* ---
Set the font
*/

void ps_set_font(FILE *fp, int newfont)
{
	int name;

	if (newfont == lastfont)
		return;
	lastfont = newfont;
	if (newfont == -1) return;	/* starter mark */
	name = font_table[newfont].name;
	fprintf(fp, "/%s%s findfont\n",
		ps_fontname(newfont),
		fontname_table[name].iso8859_1?"-ISOLatin1":"");
	fprintf(fp, "%d scalefont\n", ps_fontsize(newfont));
	fprintf(fp, "setfont\n");
}

